# Copyright 2017 TsumiNa. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from collections import Counter

import numpy as np
import pandas as pd
from scipy import stats
from matminer.featurizers.structure import RadialDistributionFunction
from sklearn.base import BaseEstimator, TransformerMixin

# from ..pipeline import combinator


class PDFDesc(BaseEstimator, TransformerMixin):
    """
    Calculate pair distribution descriptor for machine learning.

    """

    def __init__(self,
                 n_grid: int = 151,
                 r_max: float = 20.0,
                 *,
                 structure_col: str = 'structure',
                 verbose: bool = True):
        """

        Parameters
        ----------
        n_grid: int
            Number of radial grid points.
        r_max: float
            Maximum of radial grid (the minimum is always set zero).
        structure_col: str
            Column's name contain structure objects.
        verbose: bool
            If verbose output.
        """

        assert n_grid >= 1, "ngrid should be greater than 1!"

        self.structures = list()
        self.structure_col = structure_col
        self.r_max = r_max
        self.dr = r_max / (n_grid - 1)
        self.interval = np.arange(0, r_max + self.dr, self.dr)
        self.rdf = RadialDistributionFunction(r_max, self.dr)
        self.verbose = verbose
        self.pdfs = None

    def fit(self, structures):
        """

        Parameters
        ----------
        structures: Structure list
            pymatgen Structure object list.

        Returns
        -------
            self
        """
        if self.structure_col not in structures:
            raise ValueError('need structure to calculate descriptor')

        return self

    def transform(self, structures):
        if self.structure_col not in structures:
            raise ValueError('need structure to calculate descriptor')

        if self.verbose:
            import time
            start = time.time()
            print('converting {:d} structure'.format(structures.shape[0]))

            ret = structures[self.structure_col].apply(self.pdf).tolist()
            ret = pd.DataFrame(
                np.array(ret), index=structures.index, columns=self.interval)

            end = time.time()
            print('using {:.5f}s'.format(end - start))
        else:
            ret = structures[self.structure_col].apply(self.pdf).tolist()
            ret = pd.DataFrame(
                np.array(ret), index=structures.index, columns=self.interval)

        self.pdfs = ret
        return ret

    def _pdf(self, s):
        _, rdf = self.rdf(s)[0]
        pass

    def to_csv(self, file: str):
        """
        Output RDF data to a csv file.

        :param str file: file path to save.
        """
        self.pdfs.to_csv(file)



# Copyright 2017 TsumiNa. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

__version__ = '0.1.0'
__release__ = 'b4'
__short_description__ = "material descriptor library"
__license__ = "BSD (3-clause)"
__author__ = "TsumiNa"
__author_email__ = "liu.chang.1865@gmail.com"
__maintainer__ = "TsumiNa"
__maintainer_email__ = "liu.chang.1865@gmail.com"
__github_username__ = "yoshida-lab"

__pkg_cfg_root__ = '.' + __name__


def get_conf(key: str = None):
    """
    Return config value with key or all config.

    Parameters
    ----------
    key: str or None
        name of config item.
        left `None` will return config entry as dict.

    Returns
    -------
    object
        key value in ``conf.yml`` file.
    """
    import yaml
    from pathlib import Path
    home = Path.home()
    dir_ = home / __pkg_cfg_root__
    cfg_file = dir_ / 'conf.yml'
    with open(cfg_file) as f:
        conf = yaml.load(f)
    if not key:
        return conf
    else:
        return conf[key]


def _get_dataset_url(fname: str):
    """
    Return url with the given file name.

    Args
    ----
    fname: str
        binary file name.

    Return
    ------
    str
        binary file url.
    """

    return 'https://github.com/' + \
           __github_username__ + '/' + \
           __name__ + '/releases/download/' + \
           __version__ + __release__ + '/' + \
           fname + '.pkl'


def _init_cfg_file(force=False):
    """
    Create config file is not exist at ~/.xenonpy/conf.yml

    ..warning::
        Set **force=True** will reset all which under the `~/.xenonpy`` dir.

    Agrs
    ----
    force: bool
        force reset ``conf.yml`` to default and empty all dirs under ``~/.xenonpy``.
    """
    from shutil import rmtree, copyfile
    from pathlib import Path
    home = Path.home()
    dir_ = home / __pkg_cfg_root__
    cfg_file = dir_ / 'conf.yml'

    dataset_dir = dir_ / 'dataset'
    userdata_dir = dir_ / 'userdata'
    cached_dir = dir_ / 'cached'

    if force:
        rmtree(str(dir_))

    if not dir_.is_dir():
        # create config root dir
        dir_.mkdir()

        # create other dirs
        dataset_dir.mkdir()
        cached_dir.mkdir()
        userdata_dir.mkdir()

    if not cfg_file.exists() or force:
        # copy default conf.yml to ~/.xenonpy
        copyfile(str(Path(__file__).parent / 'conf.yml'), str(cfg_file))


_init_cfg_file()

from . import descriptor
from . import model
# from .pipeline import *
# from .preprocess import *
from . import utils
from . import visualization



# Copyright 2017 TsumiNa. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from pathlib import Path

import numpy as np
import torch as tc
import torch.nn as nn
from sklearn.base import BaseEstimator, RegressorMixin
from torch.autograd import Variable as V

from .checkpoint import CheckPoint
from .layer import Layer1d


# from ...pipeline import combinator


class NNGenerator1d(object):
    def __init__(self, n_features: int, n_predict: int, *,
                 n_neuron: [int],
                 p_drop: [float] = (0.0,),
                 layer_func: [] = (nn.Linear,),
                 act_func: [] = (nn.ReLU(),),
                 batch_normalize: [bool] = (False,),
                 momentum: [float] = (0.1,)
                 ):
        self.n_in, self.n_out = n_features, n_predict

        # save parameters
        self.n_neuron = n_neuron
        self.p_drop = p_drop
        self.layer_func = layer_func
        self.act_func = act_func
        self.batch_normalize = batch_normalize
        self.momentum = momentum

        # prepare layer container
        # calculate layer's variety
        self.layers = list()
        self.layers_len = 0
        self.__lay_vars()

    def __lay_vars(self):
        for n in self.n_neuron:
            for p in self.p_drop:
                for l in self.layer_func:
                    for a in self.act_func:
                        for b in self.batch_normalize:
                            for m in self.momentum:
                                layer = dict(n_in=0,
                                             n_out=n,
                                             p_drop=p,
                                             layer_func=l,
                                             act_func=a,
                                             batch_normalize=b,
                                             momentum=m)
                                self.layers.append(layer)
        self.layers_len = len(self.layers)

    def __call__(self, hidden=3):
        ids = np.random.randint(0, self.layers_len, hidden)
        layers = list()

        # set layers
        self.layers[ids[0]]['n_in'] = self.n_in
        n_in = self.n_in
        for i in ids:
            self.layers[i]['n_in'] = n_in
            layers.append(Layer1d(**self.layers[i]))
            n_in = self.layers[i]['n_out']
        out_layer = Layer1d(n_in=n_in, n_out=self.n_out, act_func=None)
        layers.append(out_layer)

        return nn.Sequential(*layers)


class ModelRunner(BaseEstimator, RegressorMixin):
    def __init__(self, model: nn.Module, *,
                 lr: float = 0.01,
                 loss_func=None,
                 optim=None,
                 epochs: int = 2000,
                 verbose: int = 0,
                 ctx: str = 'cpu',
                 checkstep: int = 100,
                 check: tuple = ('epochs', 'loss'),
                 save_to: str = ''
                 ):
        parent = Path(save_to)
        if not parent.exists():
            parent.mkdir()
        self.save_snap = str(Path(save_to) / 'snap.pkl')
        self.save_model = str(Path(save_to) / 'model.pkl')
        self.checkstep = checkstep
        self.check = check
        self.trained = False
        self.model = model
        self.checkpoint = CheckPoint(model, *check)
        self.x = None
        self.y = None
        self.lr = lr
        self.ctx = ctx
        self.loss_func = loss_func
        self.optim = optim
        self.epochs = epochs
        self.verbose = verbose

    def fit(self, x, y=None):

        # transform to torch tensor
        if not isinstance(x, V):
            x = V(x, requires_grad=False)
        _, col = x.size()
        if not isinstance(y, V):
            y = V(y, requires_grad=False)

        # optimization
        optim = self.optim(self.model.parameters(), lr=self.lr)

        # if use CUDA acc
        if self.ctx == 'GPU' and tc.cuda.is_available():
            self.model.cuda()
            x = x.cuda()
            y = y.cuda()

        # train
        loss = None
        for t in range(self.epochs):
            pre_y = self.model(x)
            loss = self.loss_func(pre_y, y)
            optim.zero_grad()
            loss.backward()
            optim.step()

            if self.verbose and t % self.verbose == 0:
                print('at step: {}, Loss={:.4f}'.format(t, loss.data[0]))
            if self.checkstep > 0 and t % self.checkstep == 0:
                self.checkpoint(epochs=t, loss=loss.data[0])

        self.trained = True
        if self.checkstep > 0:
            self.checkpoint.save(self.save_snap, self.save_model)
        print('Loss={:.4f}'.format(loss.data[0]))
        return self

    def predict(self, x, y=None):
        if not isinstance(x, V):
            x = V(x)
        pre_y = self.model(x)
        if y:
            if not isinstance(y, V):
                y = V(y)
            print(self.loss_func(pre_y, y))
        return pre_y



# Copyright 2017 TsumiNa. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sb
from scipy.stats import boxcox
from sklearn.base import BaseEstimator
from sklearn.preprocessing import minmax_scale


class DescHeatmap(BaseEstimator):
    def __init__(self,
                 save=None,
                 bc=False,
                 pivot_kws=None,
                 method='average',
                 metric='euclidean',
                 figsize=None,
                 row_cluster=False,
                 col_cluster=True,
                 row_linkage=None,
                 col_linkage=None,
                 row_colors=None,
                 col_colors=None,
                 mask=None,
                 **kwargs):
        self.save = save
        self.bc = bc
        self.lmd = None
        self.pivot_kws = pivot_kws
        self.method = method
        self.metric = metric
        self.figsize = figsize
        self.col_cluster = col_cluster
        self.row_linkage = row_linkage
        self.row_cluster = row_cluster
        self.col_linkage = col_linkage
        self.row_colors = row_colors
        self.col_colors = col_colors
        self.mask = mask
        self.kwargs = kwargs
        self.desc = None

    def _bc(self, series):
        series_bc, self.lmd = boxcox(series - series.min() + 0.01)
        return series_bc

    def fit(self, desc):
        desc_ = desc
        if self.bc:
            desc_ = desc.apply(self._bc)
        self.desc = pd.DataFrame(minmax_scale(desc_), index=desc.index, columns=desc.columns)
        return self

    def draw(self, y):
        ax = sb.clustermap(
            self.desc,
            cmap="RdBu",
            method=self.method,
            figsize=self.figsize,
            row_cluster=self.row_cluster,
            col_cluster=self.col_cluster,
            **self.kwargs)
        ax.cax.set_visible(False)
        ax.ax_heatmap.set_position((0.1, 0.2, 0.84, 0.6))
        ax.ax_heatmap.yaxis.set_ticks_position('left')
        ax.ax_heatmap.yaxis.set_label_position('left')
        ax.ax_col_dendrogram.set_position((0.1, 0.8, 0.83, 0.1))

        ax = plt.axes([0.95, 0.2, 0.05, 0.6])
        ax.plot(y.values, lw=4)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.set_xlabel('{:s}\n(property)'.format(y.name), fontsize=20)
        if self.save:
            plt.savefig(**self.save)



# Copyright 2017 TsumiNa. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import uuid
from collections import defaultdict
from datetime import datetime as dt
from os import remove
from os.path import getmtime
from pathlib import Path
from shutil import rmtree
from urllib.parse import urlparse

import pandas as pd
import requests
from sklearn.externals import joblib as jl

from .. import __pkg_cfg_root__
from .. import _get_dataset_url


class Loader(object):
    """
    Load data from embed dataset in XenonPy's or user create data saved in ``~/.xenonpy/cached`` dir.
    Also can fetch data by http request.

    This is sample to demonstration how to use is. Also see parameters documents for details.

    ::

        >>> load = Loader()
        >>> elements = load('elements')
        >>> ele.info()
        <class 'pandas.core.frame.DataFrame'>
        Index: 118 entries, H to Og
        Data columns (total 74 columns):
        atomic_number                    118 non-null int64
        atomic_radius                    88 non-null float64
        atomic_radius_rahm               96 non-null float64
        atomic_volume                    91 non-null float64
        atomic_weight                    118 non-null float64
        boiling_point                    96 non-null float64
        brinell_hardness                 59 non-null float64
        bulk_modulus                     69 non-null float64
        c6                               43 non-null float64
        c6_gb                            86 non-null float64
        covalent_radius_bragg            37 non-null float64
        covalent_radius_cordero          96 non-null float64
        covalent_radius_pyykko           118 non-null int64
        covalent_radius_pyykko_double    108 non-null float64
        covalent_radius_pyykko_triple    80 non-null float64
        covalent_radius_slater           86 non-null float64
        density                          95 non-null float64
        dipole_polarizability            106 non-null float64
        electron_negativity              103 non-null float64
        electron_affinity                77 non-null float64
        en_allen                         71 non-null float64
        en_ghosh                         103 non-null float64
        en_pauling                       85 non-null float64
        first_ion_en                     103 non-null float64
        fusion_enthalpy                  92 non-null float64
        gs_bandgap                       112 non-null float64
        gs_energy                        112 non-null float64
        gs_est_bcc_latcnt                112 non-null float64
        gs_est_fcc_latcnt                112 non-null float64
        gs_mag_moment                    112 non-null float64
        gs_volume_per                    112 non-null float64
        hhi_p                            77 non-null float64
        hhi_r                            77 non-null float64
        heat_capacity_mass               85 non-null float64
        heat_capacity_molar              85 non-null float64
        icsd_volume                      112 non-null float64
        evaporation_heat                 88 non-null float64
        gas_basicity                     32 non-null float64
        heat_of_formation                89 non-null float64
        lattice_constant                 87 non-null float64
        linear_expansion_coefficient     63 non-null float64
        mendeleev_number                 103 non-null float64
        melting_point                    100 non-null float64
        metallic_radius                  56 non-null float64
        metallic_radius_c12              63 non-null float64
        molar_volume                     97 non-null float64
        num_unfilled                     112 non-null float64
        num_valance                      112 non-null float64
        num_d_unfilled                   112 non-null float64
        num_d_valence                    112 non-null float64
        num_f_unfilled                   112 non-null float64
        num_f_valence                    112 non-null float64
        num_p_unfilled                   112 non-null float64
        num_p_valence                    112 non-null float64
        num_s_unfilled                   112 non-null float64
        num_s_valence                    112 non-null float64
        period                           118 non-null int64
        poissons_ratio                   54 non-null float64
        proton_affinity                  32 non-null float64
        specific_heat                    81 non-null float64
        thermal_conductivity             96 non-null float64
        vdw_radius                       103 non-null float64
        vdw_radius_alvarez               94 non-null float64
        vdw_radius_batsanov              65 non-null float64
        vdw_radius_bondi                 28 non-null float64
        vdw_radius_dreiding              21 non-null float64
        vdw_radius_mm3                   94 non-null float64
        vdw_radius_rt                    9 non-null float64
        vdw_radius_truhlar               16 non-null float64
        vdw_radius_uff                   103 non-null float64
        sound_velocity                   72 non-null float64
        vickers_hardness                 39 non-null float64
        Polarizability                   102 non-null float64
        youngs_modulus                   63 non-null float64
        dtypes: float64(71), int64(3)
        memory usage: 69.1+ KB
    """

    # set to check params

    def __init__(self,
                 *,
                 chunk_size: int = 256 * 1024,
                 api_key: str = None,
                 username: str = None,
                 password: str = None,
                 payload: dict = None):
        """

        Parameters
        ----------
        url: str
            Where data are.
            None(default) means to load data in ``~/.xenonpy/cached`` or ``~/.xenonpy/dataset`` dir.
            When given as url, later can be uesd to fetch files under this url from http-request.

        chunk_size: int
            Http-request chunk size.

        api_key: str
            If remote access need a api key.

        username: str
            If need user name.

        password: str
            If need password.

        payload: dict
            If need payload.

        """
        self.payload = payload
        self.pass_word = password
        self.user_name = username
        self.api_key = api_key
        self.datasets = [
            'elements', 'elements_completed', 'mp_inorganic',
            'electron_density', 'sample_A', 'mp_structure'
        ]
        self.chunk_size = chunk_size
        self.dataset_dir = Path().home() / __pkg_cfg_root__ / 'dataset'
        self.user_data_dir = Path().home() / __pkg_cfg_root__ / 'userdata'
        self.cached_dir = Path().home() / __pkg_cfg_root__ / 'cached'

    def _fetch_data(self, url):
        schemes = {'http', 'https'}
        scheme = urlparse(url).scheme
        if 'http' in scheme:
            return self._http_data(url)
        else:
            raise ValueError("Only can access [{}] data but you send {}. :(".format(schemes, scheme))

    def _http_data(self, url):
        r = requests.get(url, stream=True)
        r.raise_for_status()

        if 'filename' in r.headers:
            filename = r.headers['filename']
        else:
            filename = url.split('/')[-1]

        save_to = str(self.cached_dir / filename)
        with open(save_to, 'wb') as f:
            for chunk in r.iter_content(chunk_size=self.chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return save_to

    def __call__(self, dname: str):
        """
        load preset dataset.

        .. note::
            Try to load data from local at ``~/.xenonpy/dataset``.
            If no data, try to fetch them from remote repository.

        Args
        -----------
        dname: str
            name of dateset.

        Returns
        ------
        ret:DataFrame or Saver or local file path.
        """

        # check dataset name
        if dname in self.datasets:
            dataset = self.dataset_dir / (dname + '.pkl.pd_')

            # check dataset exist
            if not dataset.exists():
                self._fetch_data(_get_dataset_url(dname), dataset)

            # fetch data from source
            return pd.read_pickle(str(dataset))

        elif '/' not in dname:
            dataset = self.user_data_dir / dname
            if not dataset.is_dir():
                raise FileNotFoundError(
                    'no user dataset under {}'.format(dataset))
            return Saver(dname)

        else:
            file_name = self._fetch_data(dname)
            dataset = self.cached_dir / file_name
            return dataset

    @property
    def elements(self):
        """
        Element properties from embed dataset.
        These properties are summarized from `mendeleev`_, `pymatgen`_, `CRC Handbook`_ and `magpie`_.

        See Also: :doc:`dataset`

        .. _mendeleev: https://mendeleev.readthedocs.io
        .. _pymatgen: http://pymatgen.org/
        .. _CRC Handbook: http://hbcponline.com/faces/contents/ContentsSearch.xhtml
        .. _magpie: https://bitbucket.org/wolverton/magpie

        Returns
        -------
        DataFrame:
            element properties in pd.DataFrame
        """
        return self('elements')

    @property
    def elements_completed(self):
        """
        Completed element properties. [MICE]_ imputation used

        .. [MICE] `Int J Methods Psychiatr Res. 2011 Mar 1; 20(1): 40–49.`__
                    doi: `10.1002/mpr.329 <10.1002/mpr.329>`_

        .. __: https://www.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi?dbfrom=pubmed&retmode=ref&cmd=prlinks&id=21499542

        See Also: :doc:`dataset`

        Returns
        -------
            imputed element properties in pd.DataFrame
        """
        return self('elements_completed')


class Saver(object):
    """
    Save data in a convenient way:

    .. code:: python

        import numpy as np
        np.random.seed(0)

        # some data
        data1 = np.random.randn(5, 5)
        data2 = np.random.randint(5, 5)

        # init Saver
        save = Saver('you_dataset_name')

        # save data
        save(data1, data2)

        # retriever data
        date = save.last()  # last saved
        data = save[0]  # by index
        for data in save:  # as iterator
            do_something(data)

        # delete data
        save.delete(0)  # by index
        save.delete()  # delete 'you_dataset_name' dir

    See Also: :doc:`dataset`
    """

    def __init__(self, dataset=None):
        """
        Parameters
        ----------
        dataset: str
            The dir to save and load data
        """
        self.dataset = dataset
        self._path = Path.home() / __pkg_cfg_root__ / 'userdata' / dataset
        if not self._path.exists():
            self._path.mkdir(parents=True)
        self._files = None
        self._make_file_index()

    def _make_file_index(self):
        self._files = defaultdict(list)
        files = [f for f in self._path.iterdir() if f.match('*.pkl.*')]

        for f in files:
            # named data
            if len(f.suffixes) == 3:
                d_name = f.suffixes[0].lstrip('.')
                self._files[d_name].append(f)
                continue

            # unnamed data
            self._files['temp_data'].append(f)

        for v in self._files.values():
            self._sort_files(v)

    @classmethod
    def _sort_files(cls, files):
        if files is not None:
            files.sort(key=lambda f: getmtime(str(f)))

    @classmethod
    def _load_data(cls, file):
        if file.suffix == '.pd_':
            return pd.read_pickle(str(file))
        else:
            return jl.load(file)

    def _save_data(self, data, filename=None):
        uid = str(uuid.uuid1()).replace('-', '')
        prefix = uid + '.' + filename if filename else uid
        if not self._path.exists():
            self._path.mkdir()

        if isinstance(data, pd.DataFrame):
            file = self._path / (prefix + '.pkl.pd_')
            pd.to_pickle(data, str(file))
        else:
            file = self._path / (prefix + '.pkl.z')
            jl.dump(data, file)

        return file

    def dump(self, fpath: str, *,
             fname: str = None,
             with_datetime: bool = True):
        """
        Dump last checked dataset to file.

        Parameters
        ----------
        fpath: str
            Where save to.
        fname: str
            File name. Omit to use dataset name.
        with_datetime: bool
            Suffix file name with dumped time.

        Returns
        -------
        ret: str
            File path.
        """
        ret = {k: self._load_data(v[-1]) for k, v in self._files.items()}
        name = fname if fname else self.dataset
        if with_datetime:
            datetime = dt.now().strftime('-%Y-%m-%d_%H-%M-%S_%f')
        else:
            datetime = ''
        path_dir = Path(fpath).expanduser()
        if not path_dir.exists():
            path_dir.mkdir(parents=True, exist_ok=True)
        path = path_dir / (name + datetime + '.pkl.z')
        jl.dump(ret, path)

        return str(path)

    def last(self, d_name: str = None):
        """
        Return last saved data.

        Args
        ----
        d_name: str
            Data's name. Omit for access temp data

        Return
        -------
        ret:any python object
            Data stored in `*.pkl` file.
        """
        if d_name is None:
            return self._load_data(self._files['temp_data'][-1])
        return self._load_data(self._files[d_name][-1])

    def rm(self, index, d_name: str = None):
        """
        Delete file(s) with given index.

        Parameters
        ----------
        index: int or slice
            Index of data. Data sorted by datetime.
        d_name: str
            Data's name. Omit for access unnamed data.
        """

        if not d_name:
            files = self._files['temp_data'][index]
            if not isinstance(files, list):
                remove(str(files))
            else:
                for f in files:
                    remove(str(f))
            del self._files['temp_data'][index]
            return

        files = self._files[d_name][index]
        if not isinstance(files, list):
            remove(str(files))
        else:
            for f in files:
                remove(str(f))
        del self._files[d_name][index]

    def clean(self, data_name: str = None):
        """
        Remove all data by name. Omit to remove hole dataset.

        Parameters
        ----------
        data_name: str
            Data's name.Omit to remove hole dataset.
        """
        if data_name is None:
            rmtree(str(self._path))
            self._files = list()
            self._files = defaultdict(list)
            return

        for f in self._files[data_name]:
            remove(str(f))
        del self._files[data_name]

    def __repr__(self):
        cont_ls = []
        cont_ls.append('"{}" include:'.format(self.dataset))

        for k, v in self._files.items():
            cont_ls.append('"{}": {}'.format(k, len(v)))

        return '\n'.join(cont_ls)

    def __getitem__(self, item):

        # load file
        def _load_file(files, item):
            _files = files[item]
            if not isinstance(_files, list):
                return self._load_data(_files)
            return [self._load_data(f) for f in _files]

        if isinstance(item, tuple):
            try:
                key, index = item
            except ValueError:
                raise ValueError('except 2 parameters. [str, int or slice]')
            if not isinstance(key, str) or \
                    (not isinstance(index, int) and not isinstance(index, slice)):
                raise ValueError('except 2 parameters. [str, int or slice]')
            return _load_file(self._files[key], index)

        if isinstance(item, str):
            return self.__getitem__((item, slice(None, None, None)))

        return _load_file(self._files['temp_data'], item)

    def __call__(self, *data, **named_data):
        for d in data:
            f = self._save_data(d)
            self._files['temp_data'].append(f)

        for k, v in named_data.items():
            f = self._save_data(v, k)
            self._files[k].append(f)



# Copyright 2017 TsumiNa. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import numpy as np
import pandas as pd
from pymatgen import Composition
from sklearn.base import TransformerMixin, BaseEstimator

from ..utils.datatools import Loader


# from ..pipeline import combinator


class ElementDesc(BaseEstimator, TransformerMixin):
    """
    Calculate elemental descriptors from compound's composition. i.e:

        >>> from xenonpy.descriptor import ElementDesc
        >>> Desc = ElementDesc()
        >>> desc = Desc.fit_transform(compounds)
    """

    #: methods can be used to calculate descriptors.
    methods = {'weighted_average', 'weighted_sum', 'weighted_variance', 'max', 'min'}

    def __init__(self,
                 *methods: str,
                 elements=None,
                 comps_col: str = 'composition',
                 verbose: bool = True
                 ):
        """

        Parameters
        ----------
        methods: str
            Calculation method(s) which to be used must in the :attr:`methods` list.
        elements: panda.DataFrame
            Elements information in `pandas.DataFrame` object. indexed by element symbol.
        comps_col: str
            Column's name which have compound's ``composition``.
            ``composition`` described in dict object.
        verbose: bool
            If verbose output.
        """

        if methods:
            if not set(methods).issubset(self.methods):
                raise ValueError(
                    'method(s) must be as "{:s}", left it None to use all'.
                        format(', '.join(self.methods)))
            self.methods = methods

        if elements:
            self.elems = elements
        else:
            # csv_path = str(Path(__file__).parent / 'completed.csv')
            self.elems = Loader().elements_completed

        self.verbose = verbose
        self.comps_col = comps_col
        self.labs = []
        self.funcs = []

    def fit(self, _):
        raw_labs = self.elems.columns.tolist()

        def ave(nums, elems):
            w_nums = nums / np.sum(nums)
            return w_nums.dot(elems)

        def sum_(nums, elems):
            return nums.dot(elems)

        def var(nums, elems):
            w_nums = nums / np.sum(nums)
            e_mean = w_nums.dot(elems)
            cen_elems = elems - e_mean
            return nums.dot(cen_elems ** 2)

        def max_(_, elems):
            return elems.max()

        def min_(_, elems):
            return elems.min()

        if 'weighted_average' in self.methods:
            self.labs += ['ave:' + s for s in raw_labs]
            self.funcs.append(ave)
        if 'weighted_sum' in self.methods:
            self.labs += ['sum:' + s for s in raw_labs]
            self.funcs.append(sum_)
        if 'weighted_variance' in self.methods:
            self.labs += ['var:' + s for s in raw_labs]
            self.funcs.append(var)
        if 'max' in self.methods:
            self.labs += ['max:' + s for s in raw_labs]
            self.funcs.append(max_)
        if 'min' in self.methods:
            self.labs += ['min:' + s for s in raw_labs]
            self.funcs.append(min_)

        return self

    def transform(self, cpds):
        if self.comps_col not in cpds:
            raise ValueError('need composition to calculate descriptor')

        def mk_desc(comp):
            # comp = comps[0]
            elems, nums = [], []
            if isinstance(comp, Composition):
                comp = comp.as_dict()
            for e, n in comp.items():
                elems.append(e)
                nums.append(n)
            nums = np.array(nums)
            elems = self.elems.loc[elems, :]
            ret = np.array([])
            for func in self.funcs:
                ret = np.concatenate((ret, func(nums, elems)))
            return ret

        if self.verbose:
            print('converting {:d} compounds'.format(cpds.shape[0]))
        ret = cpds[self.comps_col].apply(mk_desc).tolist()
        return pd.DataFrame(np.array(ret), index=cpds.index, columns=self.labs)



# Copyright 2017 TsumiNa. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# -*- coding: utf-8 -*-

# uncomment this for compatibility with py27
# from __future__ import print_function

import os
from datetime import date

from setuptools import setup, find_packages

# --- import your package ---
import xenonpy as package


def get_requirements(filename: str):
    """
    Return requirements list from a text file.

    Parameters
    ----------
    filename: str
        Name of requirement file.

    Returns
    -------
        str-list
    """
    try:
        require = list()
        f = open(filename, "rb")
        for line in f.read().decode("utf-8").split("\n"):
            line = line.strip()
            if "#" in line:
                line = line[:line.find("#")].strip()
            if line:
                require.append(line)
    except IOError:
        print("'{}' not found!".format(filename))
        require = list()

    return require


if __name__ == "__main__":
    # --- Automatically generate setup parameters ---
    # Your package name
    PKG_NAME = package.__name__

    # Your GitHub user name
    try:
        GITHUB_USERNAME = package.__github_username__
    except:
        GITHUB_USERNAME = "Unknown-Github-Username"

    # Short description will be the description on PyPI
    try:
        SHORT_DESCRIPTION = package.__short_description__  # GitHub Short Description
    except:
        print(
            "'__short_description__' not found in '%s.__init__.py'!" % PKG_NAME)
        SHORT_DESCRIPTION = "No short description!"

    # Long description will be the body of content on PyPI page
    try:
        LONG_DESCRIPTION = open("README.rst", "rb").read().decode("utf-8")
    except:
        LONG_DESCRIPTION = "No long description!"

    # Version number, VERY IMPORTANT!
    VERSION = package.__version__ + package.__release__

    # Author and Maintainer
    try:
        AUTHOR = package.__author__
    except:
        AUTHOR = "Unknown"

    try:
        AUTHOR_EMAIL = package.__author_email__
    except:
        AUTHOR_EMAIL = None

    try:
        MAINTAINER = package.__maintainer__
    except:
        MAINTAINER = "Unknown"

    try:
        MAINTAINER_EMAIL = package.__maintainer_email__
    except:
        MAINTAINER_EMAIL = None

    PACKAGES, INCLUDE_PACKAGE_DATA, PACKAGE_DATA, PY_MODULES = (
        None, None, None, None,
    )

    # It's a directory style package
    if os.path.exists(__file__[:-8] + PKG_NAME):
        # Include all sub packages in package directory
        PACKAGES = find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"])

        # Include everything in package directory
        INCLUDE_PACKAGE_DATA = True
        PACKAGE_DATA = {
            "": ["*.*"],
        }


    # It's a single script style package
    elif os.path.exists(__file__[:-8] + PKG_NAME + ".py"):
        PY_MODULES = [PKG_NAME, ]

    # The project directory name is the GitHub repository name
    repository_name = package.__name__

    # Project Url
    URL = "https://github.com/{0}/{1}".format(GITHUB_USERNAME, repository_name)
    # Use todays date as GitHub release tag
    github_release_tag = 'v' + VERSION
    # Source code download url
    DOWNLOAD_URL = "https://github.com/{0}/{1}/archive/{2}.tar.gz".format(
        GITHUB_USERNAME, repository_name, github_release_tag)

    try:
        LICENSE = package.__license__
    except:
        print("'__license__' not found in '%s.__init__.py'!" % PKG_NAME)
        LICENSE = ""

    PLATFORMS = [
        "Windows",
        "MacOS",
        "Unix",
    ]

    CLASSIFIERS = [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ]

    # Read requirements.txt, ignore comments
    REQUIRES = get_requirements("requirements.txt")
    REQUIRES_TEST = get_requirements("requirements_test.txt")

    setup(
        name=PKG_NAME,
        description=SHORT_DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        version=VERSION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        packages=PACKAGES,
        include_package_data=INCLUDE_PACKAGE_DATA,
        package_data=PACKAGE_DATA,
        py_modules=PY_MODULES,
        url=URL,
        download_url=DOWNLOAD_URL,
        classifiers=CLASSIFIERS,
        platforms=PLATFORMS,
        license=LICENSE,
        setup_requires=['pytest-runner'],
        install_requires=REQUIRES,
        tests_require=REQUIRES_TEST,
    )

"""
Appendix
--------
classifiers: https://pypi.python.org/pypi?%3Aaction=list_classifiers
"""




# Copyright 2017 TsumiNa. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.


from .element import ElementDesc
from .pdf import PDFDesc
